#!/usr/bin/perl
#
#  fix_input_scripts_wl.pl:
#
#	Invocation:
#		fix_input_scripts_wl.pl <list of .in files>
#  or		fix_input_scripts_wl.pl
#  the second version operates on all .in files present in the current directory.
#
#	The script searches the output files for failures of the Cloudy findline()
#  function, and compiles a list of suggested updates.  These changes are applied
#  to a number of commands and any lists of lines that may appear in the input
#  script.  Files invoked from the script that may contain commands or linelists
#  are also parsed and corrected if needed.
#
#	A limited number of Cloudy commands is currently supported.  To process
#  another command, one must do two things:
#
#  1) Define two routines of the form "&match_any_cmd()" and "&match_this_cmd()"
#     (see below for examples) that evaluate true when the script line matches
#     any command of that type, or the command with the intended change has been
#     performed already (in a previous run).
#
#  2) Append the @cmds array below with references to these routines.
#
#  Chatzikos, Marios			2013-Jul-25
#

use strict;
use warnings;



my ($comment, $datestr);

my $end_of_lines = qr/^end\s+(of\s+|)line/;
my $datadir = "../../data";

#
#  These are references to functions aimed to match certain commands.  The array
#  is populated below, after the functions are defined.
#
my @cmds;


#
#  Switch on to get detailed report
#
my $verbose = 0;





#################################################################################
#										#
#				FIELD PROCESSING				#
#										#
#################################################################################

#
#  Subroutines used to identify script lines of certain commands (*_any_* subs)
#  and commands of a given transition (*_this_*).
#  These are mainly used as input to &process_line().
#

#
#  Routines to match commands....
#
sub match_this
{
	my ($script_line, $label, $wvlng) = @_;
	my $match = 0;
	$match = 1
		if( $script_line =~ /$label/i	and
		    $script_line =~ /$wvlng/i );
	return	$match;
}


sub match_any_monitor_line
{
	my ($script_line) = @_;
	my $match = 0;
	$match = 1
		if( $script_line =~ m/^moni[a-zA-Z]*/i	and
		    $script_line =~ m/line\s+/i );
	return	$match;
}


sub match_this_monitor_line
{
	my ($script_line, $label, $wvlng) = @_;
	my $match = 0;
	$match = 1
		if( &match_any_monitor_line($script_line)	and
		    &match_this($script_line, $label, $wvlng) );
	return	$match;
}



sub match_any_normalize
{
	my ($script_line) = @_;
	my $match = 0;
	$match = 1
		if( $script_line =~ m/^norm[a-zA-Z]*/ );
	return $match;
}



sub match_this_normalize
{
	my ($script_line, $label, $wvlng) = @_;
	my $match = 0;
	$match = 1
		if( &match_any_normalize($script_line)	and
		    &match_this($script_line, $label, $wvlng) );
	return $match;
}



sub match_any_stopline
{
	my ($script_line) = @_;
	my $match = 0;
	$match = 1
		if( $script_line =~ m/^stop\s+line/ );
	return $match;
}



sub match_this_stopline
{
	my ($script_line, $label, $wvlng) = @_;
	my $match = 0;
	$match = 1
		if( &match_any_stopline($script_line) and
		    &match_this($script_line, $label, $wvlng) );
	return $match;
}



#
#  Routines to match line lists....
#

sub match_no_comment_line
{
	my ($script_line) = @_;
	my $match = 0;
	$match = 1
		if( $script_line !~ m/^#/  and $script_line !~ m/^c/  and $script_line !~ m/^\%/ and
		    $script_line !~ m#^//# and $script_line !~ m/^\*/ );
	return	$match;
}



sub match_listed_line
{
	my ($script_line, $label, $wvlng) = @_;
	my $match = 0;
	$match = 1
		if( &match_no_comment_line($script_line) and
		    &match_this($script_line, $label, $wvlng) );
	return	$match;
}



sub match_any_linelist_cmd
{
	my ($script_line) = @_;
	my $match = 0;
	$match = 1
		if( $script_line =~ m/^(opti|prin|save)/ );
	return	$match;
}



sub match_end_of_lines
{
	my ($script_line) = @_;
	my $match = 0;
	$match = 1
		if( $script_line =~ m/$end_of_lines/ );
	return $match;
}



#
#  Populate the array of function references that can regexp-match a set of
#  commands.  To expand the number of commands implemented, add suitable
#  functions above, and enter references to them at the end of the array.
#
#  NOTE:  These all should refer to commands.  Lists of lines are treated
#  separately.
#
$cmds[0]{any}  = \&match_any_monitor_line;
$cmds[0]{this} = \&match_this_monitor_line;
$cmds[1]{any}  = \&match_any_normalize;
$cmds[1]{this} = \&match_this_normalize;
$cmds[2]{any}  = \&match_any_stopline;
$cmds[2]{this} = \&match_this_stopline;



#
#  The parser may read emission lines in two ways:  As arguments to commands, in
#  which case the species is given between double quotations somewhere in the line,
#  typically followed by the wavelength.  Or as part of a list of emission lines,
#  in which case, the species is given without quotations in columns 1-4 followed
#  by the wavelength.
#
#  Lists can appear in independent files, or in the input script terminated by an
#  "end (of) lines" command.
#
#  The 3 subs below deal with properly reading and reporting the species and line
#  wavelength.
#

#
#  Emulate the parser, and use the first number in the command (possibly ending in
#  a single character unit) as the wavelength of the transition.
#
sub get_quoted_emline
{
	my ($script_line) = @_;

	my @words = split(/\"/, $script_line);

	my $scrpt_label = $words[1];
	$scrpt_label = "\Q$scrpt_label\E";

	my $scrpt_wl;
	@words = split( /\s+/, $script_line );
	foreach my $word ( @words )
	{
		if ($word =~ m/^\d+(\.\d*|)([a-zA-Z]|)$/ )
		{
			$scrpt_wl = $word;
			last;
		}
	}

	return	($scrpt_label, $scrpt_wl);
}



# Parse lines of the form
#	"H  1  1216A come what may"
sub get_listed_emline
{
	my( $script_line ) = @_;

	my ($label) = substr($script_line, 0, 4, " " x 4);
	$label = "\Q$label\E";
	my (undef, $oldwl, $rest) = split(/\s+/, $script_line, 3);

	#	print "script_line:\t \"$script_line\"\n";
	#	print "label= \"$label\"\t oldwl= $oldwl\n";

	return	($label, $oldwl, $rest);
}



sub get_emission_line
{
	my ($script_line, $quoted) = @_;

	if( defined($quoted) and $quoted )
	{
		return	&get_quoted_emline($script_line);
	}
	else
	{
		return	&get_listed_emline($script_line);
	}
}



#
#  Conversion factor for wavelength comparison.  This is required to match the
#  intended wavelength reported by Cloudy with the one that appears in the script
#  itself.
#  The factor is meant to be applied to the first of the input wavelengths.
#
sub convert_wl
{
	my ($wl_unit1, $wl_unit2) = @_;


	my $conv_fact = 1;
	if( $wl_unit2 ne $wl_unit1 )
	{
		if( $wl_unit2 eq "A" or $wl_unit2 eq "")
		{
			$conv_fact = 1e4	if( $wl_unit1 eq "m" );
			$conv_fact = 1e6	if( $wl_unit1 eq "c" );
		}
		elsif( $wl_unit2 eq "m" )
		{
			$conv_fact = 1e-4	if( $wl_unit1 eq "A" or $wl_unit1 eq "" );
			$conv_fact = 1e2	if( $wl_unit1 eq "c" );
		}
		elsif( $wl_unit2 eq "c" )
		{
			$conv_fact = 1e-6	if( $wl_unit1 eq "A" or $wl_unit1 eq "" );
			$conv_fact = 1e-2	if( $wl_unit1 eq "m" );
		}
	}
	#	print "$wl_unit2\t  VS  \t$unit_oldwl\t $conv_fact:\t $scrpt_wl\t VS ".	($oldwl * $conv_fact)	."\n";

	return	$conv_fact;
}



#
#  Process wavelength to report the numerical value, the unit, and the number of
#  significant figures found.
#
sub wl_proc
{
	my ($wl) = @_;

	my $wl_unit = "";
	$wl_unit = substr($wl, length($wl)-1, 1, "")	if ($wl =~ m/[a-zA-Z]$/);

	my $idot = index( $wl, "." );
	my $nsigfig = 0;
	if( $idot >= 0 )
	{
		$nsigfig = length(substr($wl, $idot+1));
	}

	return	($wl, $wl_unit, $nsigfig);
}



#
#  Given two wavelengths (including units), produce a string of the first
#  wavelength that best matches the format of the second wavelength.  This is
#  used to pattern match the wavelength reported by Cloudy against the one
#  that exists in the script (likely of lower precision).
#
sub wl_match
{
	my ($wl1, $wl2) = @_;

	my ($wl1str, $wl1_unit, $wl1_nsigfig) = &wl_proc($wl1);
	my ($wl2str, $wl2_unit, $wl2_nsigfig) = &wl_proc($wl2);
	my $conv_fact = &convert_wl($wl1_unit, $wl2_unit);

	return	sprintf("%.*f", $wl2_nsigfig, $wl1str * $conv_fact);
}



#
#  Report if two wavelengths match after taking into account differences in
#  units and precision.
#
sub do_wl_match
{
	my ($wl1, $wl2) = @_;

	my (undef, $wl2_unit) = &wl_proc($wl2);
	my $match_wl = &wl_match($wl1, $wl2);

	my $match = 0;
	$match = 1	if( $wl2 eq "$match_wl$wl2_unit" );
	#	print "$wl1 \t $wl2\t $match_wl\t $match\n";

	return	($match);
}



#
#  Confirm that the new wl is within Cloudy's standard error (1e-4).
#
sub wldiff_error
{
	my ($oldwl, $newwl) = @_;

	my ($wl1, $wl1_unit) = &wl_proc($oldwl);
	my ($wl2, $wl2_unit) = &wl_proc($newwl);
	my $conv_fact = &convert_wl($wl1_unit, $wl2_unit);

	$wl2 *= $conv_fact;

	my $tol = int( log($wl1) / log(10.0) );
	$tol = 0.5 * 10.0**($tol-4.0);

	my $error = abs($wl1 - $conv_fact*$wl2);

	return	($tol, $error);
}



#
#  Form the new command with the correct wavelength information, and prepend it
#  by a suitable comment line detailing the change.
#
sub fix_line
{
	my ($is_llst, $script_line, $scrpt_wl, $newwl) = @_;

	my $sentinel = "//";
	   $sentinel = "#"	if( $is_llst );

	my $comment_line = "$sentinel >>chng $datestr, wl from $scrpt_wl to $newwl, $comment\n";

	my ($prefix, $suffix) = split(/$scrpt_wl/, $script_line);
	my $new_line = $prefix . $newwl;

	$suffix =~ s/^[a-zA-z]//	if( $suffix =~ m/^[a-zA-Z]/ );
	$new_line .= $suffix;

	#	print "$script_line";
	#	print "$new_line\n\n";

	return	$comment_line . $new_line;
}



#
#  Process an input line and, if needed, replace the old wavelength.  If a match
#  is found, store the matching line for the sake of error checking at the end.
#
#  The subroutine reference $match_this_line checks against the possibility that
#  the change was done in a previous run.  One of the &match_this_*() functions
#  above should be used for that.
#
#  The parameter $is_lab_quoted distinguishes between input in a command line vs
#  in a list of lines.
#
sub process_line
{
	my ($is_llst, $script_line, $match_this_line, $is_lab_quoted, @fix_list) = @_;

	#	print "SCRIPT LINE:\t $script_line";
	my ($scrpt_label, $scrpt_wl) = &get_emission_line( $script_line, $is_lab_quoted );
	#	print "SCRIPT LINE:\t quoted= $is_lab_quoted\t label= \"$scrpt_label\"\t oldwl= \"$scrpt_wl\"\n";

	my ($need_fix) = (0, -1);
	my ($newwl, $match_old) = (0, 0);
	for( my $i = 0; $i < scalar(@fix_list); $i++ )
	{
		my $label = "\Q$fix_list[$i]{label}\E";
		#	print "\tlabel=\"$label\"\t scrpt_label=\"$scrpt_label\"\n";
		next
			if( "\U$label\E" ne "\U$scrpt_label\E" ); 

		$newwl = $fix_list[$i]{newwl};

		# Change done in a previous run
		if( &{ $match_this_line } ($script_line, $label, $newwl) )
		{
			$fix_list[$i]{found}++;
			push ( @{ $fix_list[$i]{matches} }, $script_line );
			last;
		}

		my $oldwl = $fix_list[$i]{oldwl};
		#	print "\t i= $i\t label = \"$label\"\t oldwl = $oldwl\t newwl = $newwl\n";

		if( &do_wl_match($oldwl, $scrpt_wl) )
		{
			$need_fix = 1;
			$fix_list[$i]{found}++;
			push ( @{ $fix_list[$i]{matches} }, $script_line );
			#	print "$script_line\t label= $label\t oldwl= $oldwl\t index= $index\t newwl= $newwl\n";
			#	print "=>\tslab = $scrpt_label\t swl = $scrpt_wl\t --VS-- \t lab = $label\t oldwl = $oldwl\t newwl = $newwl\t WITH\t match_old = $match_old\n\n";
			last;
		}
	}

	if( $need_fix )
	{
		$script_line = &fix_line($is_llst, $script_line, $scrpt_wl, $newwl);
	}
	#	print "FINAL SCRIPT LINE:\t $script_line\n";

	return	($script_line, @fix_list);
}





#################################################################################
#										#
#			ARRAY & FILE MANIPULATION				#
#										#
#################################################################################

sub show_array_of_fixes
{
	my (@fixes) = @_;

	for( my $i = 0; $i < scalar(@fixes); $i++)
	{
		printf "\ti=%3d\t label=\"%s\"\t oldwl=\"%s\"\t newwl=\"%s\"\n",
			$i + 1,
			$fixes[$i]{label},
			$fixes[$i]{oldwl},
			$fixes[$i]{newwl};
	}
	print	"\n";

	return;
}



sub get_file_contents
{
	my ($script) = @_;

	open SCRIPT, "< $script" or die  "Could not open:\t $script\n";
	my @contents = <SCRIPT>;
	close SCRIPT		 or print "Could not close:\t $script\n";

	return	@contents;
}



sub write_file_contents
{
	my ($fname, @newcontents) = @_;

	$fname .= ".new";

	open  NEWFILE, "> $fname" or die  "Could not open tmp sim:\t $fname\n";
	print NEWFILE @newcontents;
	close NEWFILE		 or print "Could not close tmp sim:\t $fname\n";

	return	$fname;
}



#
# Get filenames that appear between quotes as part of a command.  Some commands
# expect an input filename after the output filename.  The $offset parameter
# allows picking out the intended one.
#
sub get_aux_files
{
	my ($string, $offset, @contents) = @_;

	my @files = grep(/^$string/, @contents);

	my (@local_files, @remdir_files);
	foreach my $file (@files)
	{
		my @words = split(/\"/, $file);

		$file = undef;
		$file = $words[$offset]
			if( defined($words[$offset]) );

		next
			if( not defined( $file ) );

		if( -s $file )
		{
			push( @local_files, $file );
		}
		else
		{
			push( @remdir_files, "$datadir/$file" )
				if( -s "$datadir/$file" );
		}
	}

	return	(\@local_files, \@remdir_files);
}



sub is_hash_entered
{
	my ($hval, $key, @ahash) = @_;

	my $is_there = 0;

	for( my $i = 0; $i < scalar(@ahash); $i++)
	{
		if( $hval eq $ahash[$i]{$key} )
		{
			$is_there = 1;
			last;
		}
	}

	return	$is_there;
}



sub join_arrays
{
	my ($arr1, $arr2) = @_;
	my @arr1 = @{ $arr1 };
	my @arr2 = @{ $arr2 };

	for( my $i = 0; $i < scalar(@arr2); $i++)
	{
		if( not grep( $arr2[$i], @arr1 ) )
		{
			push( @arr1, $arr2[$i] );
		}
	}

	return	(@arr1);
}





#################################################################################
#										#
#				SCRIPT PROCESSING				#
#										#
#################################################################################

#
#  Collect the unique findline failures from the main Cloudy output file.
#
sub get_findline_fails
{
	my ($script, $output) = @_;

	my @contents = &get_file_contents( $output );
	my @fix_list;

	for( my $i = 0; $i < scalar(@contents); $i++ )
	{
		my $line = $contents[$i];

		if( $line =~ m/PROBLEM findline did not find line with label/ )
		{
			my %fix;

			my @words = split(/\"/, $line);
			$fix{label} = $words[1];
			@words      = split(/\s+/, $words[2]);
			$fix{oldwl} = $words[3];
			$fix{oldwl} =~ s/\.$//;


			++$i;
			$line = $contents[++$i];

			if( $line !~ m/The closest with correct label was/ )
			{
				print	"Script: $script\t WARNING!  No closest match found! Ignore line:\t" . $fix{label} ." @".$fix{oldwl} ."\n";
				next;
			}

			@words	    = split(/\"/, $line);
			@words	    = split(/\s+/, $words[1]);
			$fix{newwl} = $words[$#words];


			# Needed for error checking...
			$fix{found} = 0;
			$fix{matches} = [];


			#	print "label= ". $fix{label} ."\t oldwl = ". $fix{oldwl} ."\t ". $fix{newwl} ."\n";
			
			my $is_listed = 0;
			for( my $i = 0; $i < scalar(@fix_list); $i++ )
			{
				if( $fix_list[$i]{label} eq $fix{label}  and
				    &do_wl_match($fix_list[$i]{oldwl}, $fix{oldwl}) and
				    &do_wl_match($fix_list[$i]{newwl}, $fix{newwl}) )
				{
					$is_listed++;
					last;
				}
			}



			if( not $is_listed )
			{
				my ($tol, $error) = &wldiff_error( $fix{oldwl}, $fix{newwl} );
				if( $error > $tol and $verbose )
				{
					print	"\tScript: $script\t WARNING!\t Large WL diff:\t";
					printf	"\tlabel= %s\t oldwl= %s\t newwl= %s\t error= %.4f vs tol= %.4f\n",
						$fix{label}, $fix{oldwl}, $fix{newwl}, $error, $tol;
					#	print	"\tWARNING! Double check line this line.\n";
				}
				push( @fix_list, { %fix } );
			}
		}
	}

	return	@fix_list;
}



#
#  Problems with disambiguation:
#  In some cases, "H  1" lines and their "Ca B" equivalents have
#  different recommended wavelength fixes.
#  Bring these to the user's attention, and change the "Ca B" new
#  wavelength to be in agreement with the "H  1" one.
#
sub process_H1_caseb
{
	my (@fix_list) = @_;

	my $ncab = 0;
	for( my $i = 0; $i < scalar(@fix_list); $i++ )
	{
		next
			if( $fix_list[$i]{label} !~ m/^Ca (A|B)$/i );

		for( my $j = 0; $j < scalar(@fix_list); $j++ )
		{
			if( $fix_list[$j]{label} =~ m/^H  1$/i	and
				&do_wl_match($fix_list[$j]{oldwl}, $fix_list[$i]{oldwl})	and
			    not &do_wl_match($fix_list[$j]{newwl}, $fix_list[$i]{newwl}) )
			{
				print	"\t  => Setting new wl for:  "
				  .	"\"". $fix_list[$i]{label} ."\" @". $fix_list[$i]{oldwl} .","
				  .	"\t from: "             . $fix_list[$i]{newwl}
				  .	"\t   to: "             . $fix_list[$j]{newwl}
				  .	"\t to match with \""   . $fix_list[$j]{label} ."\"\n";
				$fix_list[$i]{newwl} = $fix_list[$j]{newwl};
				$ncab = 1;
				last;
			}
		}
	}

	print "\n"
		if( $ncab > 0 );
	
	return  @fix_list;
}



#
#  Loop over and fix an array of lines.  The pattern matching sub reference
#  $match_any_line should be one of the &match_any_*() functions above.
#
sub fix_contents
{
	my ($is_llst, $contents_ref, $fix_ref, $match_any_line, $match_this_line, $is_lab_quoted) = @_;
	my @contents = @{ $contents_ref };
	my @fix_list = @{ $fix_ref };

	my @newcontents;
	foreach my $script_line ( @contents )
	{
		if( &{ $match_any_line } ($script_line) )
		{
			($script_line, @fix_list) = &process_line( $is_llst, $script_line, $match_this_line, $is_lab_quoted, @fix_list );
		}

		push( @newcontents, $script_line );
	}

	return	(\@newcontents, \@fix_list);
}



#
#  Invoke the general &fix_contents() routine to carry out the wavelength
#  updates for the command specified.
#
sub fix_commands
{
	my ($is_llst, $content_ref, $fix_list_ref, $match_any_cmd, $match_this_cmd) = @_;

	my $is_lab_quoted = 1;

	my $newcont_ref;
	($newcont_ref, $fix_list_ref) =
		&fix_contents($is_llst, $content_ref, $fix_list_ref, $match_any_cmd, $match_this_cmd, $is_lab_quoted);

	return	($newcont_ref, $fix_list_ref);
}



#
#  Fix list of lines...
#
sub fix_lines_listed
{
	my ($is_llst, $content_ref, $fix_list_ref) = @_;

	my $is_lab_quoted = 0;

	my $newcont_ref;
	($newcont_ref, $fix_list_ref) =
		&fix_contents($is_llst, $content_ref, $fix_list_ref, \&match_no_comment_line, \&match_listed_line, $is_lab_quoted);

	return	($newcont_ref, $fix_list_ref);
}



#
#  Fix groups of lines in an input script, terminated by 'end (of) lines'.
#
sub fix_script_linelist
{
	my ($is_llst, $contents_ref, $fix_list_ref) = @_;
	my @contents = @{ $contents_ref };


	my $start = 0;
	my @newcontents;
	for( my $i = 0; $i < scalar(@contents); $i++ )
	{
		my $script_line = $contents[ $i ];
		if( $script_line =~ m/$end_of_lines/ )
		{
			my $end = --$i;
			do { $end--; } while( $end >= 0 and not &match_any_linelist_cmd( $contents[$end] ) );
			my @tmp = @contents[ ++$end .. $i ];
			($contents_ref, $fix_list_ref) = &fix_lines_listed($is_llst, \@tmp, $fix_list_ref);
			push( @newcontents, @contents[ $start..($end-1)] );
			push( @newcontents, @{ $contents_ref } );
			push( @newcontents, $contents[++$i]    );
			$start = $i+1;
			next;
		}
	}

	push( @newcontents, @contents[ $start..$#contents ] );

	return	(\@newcontents, $fix_list_ref);
}



#
#  Process a script of commands, that may also contain lists of lines terminated
#  by the "end of lines" sentinel, to update wavelengths to the values suggested
#  by Cloudy.  The commands processed are listed in the @cms array of hashes, see
#  above.
#
sub fix_script
{
	my (undef, $contents_ref, $fix_list_ref) = @_;

	my $is_llst = 0;

	#  Commands
	#
	for( my $icmd = 0; $icmd < scalar(@cmds); $icmd++ )
	{
		my %cmd = %{ $cmds[$icmd] };
		if( grep { &{ $cmd{any} } ( $_ ) }  @{ $contents_ref } )
		{
			($contents_ref, $fix_list_ref) = &fix_commands( $is_llst, $contents_ref, $fix_list_ref, $cmd{any}, $cmd{this} );
		}
	}

	#  Lists of lines
	#
	if( grep { &match_end_of_lines( $_ ) } @{ $contents_ref } )
	{
		($contents_ref, $fix_list_ref) = &fix_script_linelist( $is_llst, $contents_ref, $fix_list_ref );
	}


	return	($contents_ref, $fix_list_ref);
}



#
#  Apply a fix on a list of files.  The kind of fix is determined by the
#  $file_fix_sub sub-ref.  This is useful for treating external files, such as
#  those found in 'init' and 'save line' commands.  The parameter $is_llst
#  determines whether the comment sentinel: '//' for scripts, and '#' for line
#  lists.
#
sub fix_filelist
{
	my ($is_llst, $llist_ref, $fix_list_ref, $file_fix_sub) = @_;
	my @llist    = @{ $llist_ref    };

	my (@llists, $content_ref, %fix);
	foreach my $llist ( @llist )
	{
		next
			if( &is_hash_entered($llist, "file", @llists) );

		my @contents = &get_file_contents( $llist );

		($content_ref, $fix_list_ref) = &{ $file_fix_sub }( $is_llst, \@contents, $fix_list_ref );
		@contents = @{ $content_ref };

		$fix{file} = $llist;
		$fix{newfile} = &write_file_contents( $llist, @contents );

		push( @llists, { %fix } );
	}

	return	(\@llists, $fix_list_ref);
}



#
#  Final reporting...
#
sub check_fixed_wl
{
	my (@fix_list) = @_;

	my $all_found = 0;
	for (my $i = 0; $i < scalar(@fix_list); $i++)
	{
		my $found = $fix_list[$i]{found};
		my $newwl = $fix_list[$i]{newwl};

		if( $found == 1 )
		{
			$all_found++;
			splice(@fix_list, $i, 1);
			$i--;
			next;
		}
		elsif( $found == 0 )
		{
			print	"\t No matches for line:\t"
			  .	"\"". $fix_list[$i]{label} ."\" @". $fix_list[$i]{oldwl} ."\n";
		}
		elsif( $found > 1 )
		{
			print "\t WARNING:\t Multiple matches for line:\t"
			  .	"\"". $fix_list[$i]{label} ."\" @". $fix_list[$i]{oldwl} ."\n";

			my @matches = @{ $fix_list[$i]{matches} };
			for (my $i = 0; $i < scalar(@matches); $i++)
			{
				print "\t\t =>\t ". $matches[$i]
					if( $matches[$i] !~ m/$newwl/ );
			}
			print "\n";
			$all_found++;

			splice(@fix_list, $i, 1);
			$i--;
		}
	}

	return	@fix_list;
}





#################################################################################
#										#
#				MAIN PROGRAM					#
#										#
#################################################################################

print "please give a short text to be appended to the new comments in the input scripts: ";
$comment = <STDIN>;
chomp( $comment );
print	"\n";

# get the current date
my @DD = localtime;
my @abbr = qw( jan feb mar apr may jun jul aug sep oct nov dec );
$DD[5] %= 100;
$datestr = sprintf( "%2.2d %s %2.2d", $DD[5], $abbr[$DD[4]], $DD[3] );


my @scripts = @ARGV;
@scripts = glob "*.in"	if( scalar(@scripts) == 0 );

foreach my $script ( @scripts )
{
#	$script = "dynamics_veryfast.in";

	if( $script !~ m/\.in$/ )
	{
		warn	"WARNING: Not an input script:\t $script\n";
		next;
	}

	my $output = $script;
	$output =~ s/\.in$/.out/;

	next
		if( not -s $output );

	if( $verbose )
	{
		print "\n". "$script :\n";
		print "=" x length("$script :") ."\n";
	}


	my @fix_list = &get_findline_fails( $script, $output );
	@fix_list = &process_H1_caseb( @fix_list );
	&show_array_of_fixes( @fix_list )
		if( $verbose );


	my @contents = &get_file_contents( $script );
	my ($contents, $fix_list_ref) = &fix_script( undef, \@contents, \@fix_list );
	my $newscript = &write_file_contents( $script, @$contents );


	#
	#  Process auxiliary files
	#

	#
	#  External files that contain commands
	#
	my ($local_ref, $remdir_ref) = &get_aux_files("init", 1, @$contents);

	($local_ref, $fix_list_ref) = &fix_filelist( 0, $local_ref, $fix_list_ref, \&fix_script );
	my @local_inits = @{ $local_ref };

	($local_ref, $fix_list_ref) = &fix_filelist( 0, $remdir_ref, $fix_list_ref, \&fix_script );
	my @remdir_inits = @{ $local_ref };


	#
	#  External files that contain lists of lines
	#
	($local_ref, $remdir_ref) = &get_aux_files(qr/table\s+line/, 1, @$contents);
	my @local_tline  = @{ $local_ref  };
	my @remdir_tline = @{ $remdir_ref };

	($local_ref, $remdir_ref) = &get_aux_files(qr/save\s+line/, 3, @$contents);
	my @local_llist  = @{ $local_ref  };
	my @remdir_llist = @{ $remdir_ref };

	# Join lists of linelists
	#
	@local_llist  = &join_arrays(\@local_llist , \@local_tline );
	@remdir_llist = &join_arrays(\@remdir_llist, \@remdir_tline);


	my (@llists, @remdir_llists);
	($local_ref, $fix_list_ref) = &fix_filelist( 1, \@local_llist, $fix_list_ref, \&fix_lines_listed );
	@llists = @{ $local_ref };

	($local_ref, $fix_list_ref) = &fix_filelist( 1, \@remdir_llist, $fix_list_ref, \&fix_lines_listed );
	@remdir_llists = @{ $local_ref };


	#
	#  Done
	#
	@fix_list = &check_fixed_wl( @{ $fix_list_ref } );
	print	"\t After ". @remdir_inits ." non-local inits and ". @remdir_llists ." non-local linelists, unmatched corrections:\t ". @fix_list ."\n"
		if( $verbose );

	if( $verbose )
	{
		if( @fix_list )
		{
			print "\n\t WARNING: Could not match ". scalar(@fix_list) ." lines.\n";
			print "\t Inspect scripts and init/linelists:\n";
			print "\t     SCRIPT:\t $script \t  VS  \t $newscript\n";

			for( my $i = 0; $i < scalar(@local_inits); $i++ )
			{
				print "\t      INIT:\t ". $local_inits[$i]{file} ."\t  VS\t". $local_inits[$i]{newfile} ."\n";
			}
			for( my $i = 0; $i < scalar(@remdir_inits); $i++ )
			{
				print "\t DATA-INIT:\t ". $remdir_inits[$i]{file} ."\t  VS\t". $remdir_inits[$i]{newfile} ."\n";
			}
			for( my $i = 0; $i < scalar(@llists); $i++ )
			{
				print "\t      LLIST:\t ". $llists[$i]{file} ."\t  VS\t". $llists[$i]{newfile} ."\n";
			}
			for( my $i = 0; $i < scalar(@remdir_llists); $i++ )
			{
				print "\t DATA-LLIST:\t ". $remdir_llists[$i]{file} ."\t  VS\t". $remdir_llists[$i]{newfile} ."\n";
			}
		}
		else
		{
			rename "$newscript", "$script";
			for( my $i = 0; $i < scalar(@local_inits); $i++ )
			{
				print "\t init: ". $local_inits[$i]{file} ."\n";
				rename	$local_inits[$i]{newfile}, $local_inits[$i]{file};
			}
			for( my $i = 0; $i < scalar(@remdir_inits); $i++ )
			{
				print "\t remdir_inits: ". $remdir_inits[$i]{file} ."\n";
				rename	$remdir_inits[$i]{newfile}, $remdir_inits[$i]{file};
			}
			for( my $i = 0; $i < scalar(@llists); $i++ )
			{
				print "\t llist: ". $llists[$i]{file} ."\n";
				rename	$llists[$i]{newfile}, $llists[$i]{file};
			}
			for( my $i = 0; $i < scalar(@remdir_llists); $i++ )
			{
				print "\t remdir_llists: ". $remdir_llists[$i]{file} ."\n";
				rename	$remdir_llists[$i]{newfile}, $remdir_llists[$i]{file};
			}
		}
	}
	else
	{
		print "\t Script: $script\t unmatched wavelengths:\t ". @fix_list ."\n"
			if( @fix_list );
	}
#	die;
}
